// このコードに書いてあることを一から全部書くとなると骨が折れると
// 思われるかもしれませんが、一度書いてしまえば再利用できますので
// 自分の利用方法にあったツールを書くとよいかと思います
//
// 例えば改行込みのコードを出力するのに、このサンプルのように
// 毎回 `fmt.Fprintf` を使うのはとても面倒です。それらをラップ
// したツールを作るとよいでしょう。
//
// 筆者は以下のようなツールをコード生成用に開発・運用しています
//
//   * https://github.com/lestrrat-go/codegen
//   * https://github.com/lestrrat-go/xstrings
//
// これらが実際に使われているところは https://github.com/lestrrat-go/jwx
// をご覧ください。

package main

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"

	"golang.org/x/tools/imports"
)

func main() {
	if err := _main(); err != nil {
		fmt.Fprintf(os.Stderr, "%s\n", err)
		os.Exit(1)
	}
}

type Field struct {
	Name   string
	Type   string
	Getter string
}

type Object struct {
	Filename string
	Name     string
	Fields   []Field
}

func _main() error {
	// オブジェクトの定義
	objects := []Object{
		{
			Filename: `object1_gen.go`,
			Name:     `Object1`,
			Fields: []Field{
				{
					Name:   `foo`,
					Type:   `string`,
					Getter: `Foo`,
				},
				{
					Name:   `bar`,
					Type:   `int`,
					Getter: `Bar`,
				},
				{
					Name:   `baz`,
					Type:   `bool`,
					Getter: `Baz`,
				},
			},
		},
		{
			Filename: `object2_gen.go`,
			Name:     `Object2`,
			Fields: []Field{
				{
					Name:   `hoge`,
					Type:   `string`,
					Getter: `Hoge`,
				},
				{
					Name:   `fuga`,
					Type:   `int`,
					Getter: `Fuga`,
				},
				{
					Name:   `mogo`,
					Type:   `bool`,
					Getter: `Mogo`,
				},
			},
		},
	}

	for _, object := range objects {
		var buf bytes.Buffer
		fmt.Fprintf(&buf, "package article")

		fmt.Fprintf(&buf, "\n\ntype %s struct {", object.Name)
		for _, field := range object.Fields {
			fmt.Fprintf(&buf, "\n%[1]s %[2]s `json:\"%[1]s\"`", field.Name, field.Type)
		}
		fmt.Fprintf(&buf, "\n}")

		// ここではGetterを明示的に記入していますが、本来であれば
		// ExportedName()とUnexprtedName()のようなユーティリティ関数を
		// 作ってフィールド名から自動生成するとよいでしょう。
		for _, field := range object.Fields {
			fmt.Fprintf(&buf, "\n\nfunc (obj *%s) %s() %s {", object.Name, field.Getter, field.Type)
			fmt.Fprintf(&buf, "\nreturn obj.%s", field.Name)
			fmt.Fprintf(&buf, "\n}")
		}

		fmt.Fprintf(&buf, "\n\nfunc (obj *%s) UnmarshalJSON(data []byte) error {", object.Name)
		fmt.Fprintf(&buf, "\ndec := json.NewDecoder(bytes.NewReader(data))")

		fmt.Fprintf(&buf, "\n\ntok, err := dec.Token()")
		fmt.Fprintf(&buf, "\nif err != nil {")
		fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`failed to retrieve first token: %%w`, err)")
		fmt.Fprintf(&buf, "\n}") // end if err != nil

		fmt.Fprintf(&buf, "\n\ndelim, ok := tok.(json.Delim)")
		fmt.Fprintf(&buf, "\nif !ok || delim != '{' {")
		fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`expected first token to be '{' (got %%s)`, tok)")
		fmt.Fprintf(&buf, "\n}") // end if !ok || delim != '{' {

		fmt.Fprintf(&buf, "\n\nLOOP:")
		fmt.Fprintf(&buf, "\nfor {")
		fmt.Fprintf(&buf, "\ntok, err := dec.Token()")
		fmt.Fprintf(&buf, "\nif err != nil {")
		fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`failed to retrieve object key token: %%w`, err)")
		fmt.Fprintf(&buf, "\n}") // end if err != nil

		fmt.Fprintf(&buf, "\n\nif delim, ok := tok.(json.Delim); ok {")
		fmt.Fprintf(&buf, "\nif delim == '}' {")
		fmt.Fprintf(&buf, "\nbreak LOOP")
		fmt.Fprintf(&buf, "\n}") // end if delim == '}'
		fmt.Fprintf(&buf, "\n}") // end if delim, ok := ...

		fmt.Fprintf(&buf, "\n\nkey, ok := tok.(string)")
		fmt.Fprintf(&buf, "\nif !ok {")
		fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`expected object key, but got %%T`, tok)")
		fmt.Fprintf(&buf, "\n}") // end if !ok

		fmt.Fprintf(&buf, "\n\nswitch key {")
		for _, field := range object.Fields {
			fmt.Fprintf(&buf, "\ncase %q:", field.Name)
			// こんな小難しい処理をする必要はありませんが、どうやってオブジェクトの
			// 値をjson.Decoderから取ってくるのかのイメージ共有のためにあえて
			// 汎用性がある方法を紹介しています
			fmt.Fprintf(&buf, "\nvar v %s", field.Type)
			fmt.Fprintf(&buf, "\nif err := dec.Decode(&v); err != nil {")
			fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`failed to decode value for %%q: %%w`, key, err)")
			fmt.Fprintf(&buf, "\n}")
			fmt.Fprintf(&buf, "\nobj.%s = v", field.Name)
		}
		fmt.Fprintf(&buf, "\ndefault:")
		fmt.Fprintf(&buf, "\nreturn fmt.Errorf(`invalid object key %%q`, key)")
		fmt.Fprintf(&buf, "\n}") // end switch key

		fmt.Fprintf(&buf, "\n}") // end for {
		fmt.Fprintf(&buf, "\nreturn nil")
		fmt.Fprintf(&buf, "\n}")

		raw, err := ioutil.ReadAll(&buf)
		if err != nil {
			return fmt.Errorf(`failed to read source code: %w`, err)
		}

		formatted, err := imports.Process("", raw, nil)
		if err != nil {
			return fmt.Errorf(`failed to format code: %w`, err)
		}

		dst, err := os.Create(object.Filename)
		if err != nil {
			return fmt.Errorf(`failed to open %q for writing:%w`, object.Filename, err)
		}

		dst.Write(formatted)
	}
	return nil
}
